/*
 * Copyright (c) 2003 Matteo Frigo
 * Copyright (c) 2003 Massachusetts Institute of Technology
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* Do an MDCT arbitrary *even* size by a pair of FFTW R2HC (real-input
   FFTs) of half the size, plus some pre/post-processing.  This is
   based on the DCT-IV code from FFTW.  It uses a trick from:

   Zhongde Wang, "On computing the discrete Fourier and cosine transforms,"
   IEEE Trans. Acoust. Speech Sig. Proc. ASSP-33 (4), 1341--1344 (1985).

   to re-express as a pair of half-size DCT-III problems.  Our
   implementation looks quite a bit different from the algorithm
   described in the paper because we combined the paper's
   pre/post-processing with the pre/post-processing used to turn
   DCT-III into R2HC.  (Also, the paper uses a DCT/DST pair, but we
   turn the DST into a DCT via the usual reordering/sign-flip trick.
   We additionally combined a couple of the matrices/transformations
   of the paper into a single pass.)

   NOTE: We originally used a simpler method by S. C. Chan and K. L. Ho
   that turned out to have numerical problems; see reodft11e-r2hc.c
   in th FFTW 3 distribution.

   (For odd sizes, the MDCT is based on DCT-III, not DCT-IV.)
*/

#include <stdlib.h>

#include "mdct.h"

#define CONCAT(prefix, name) prefix ## name
#if defined(FFTW_SINGLE)
typedef float R;
#  define X(name) FFTW_MANGLE_FLOAT(name)
#elif defined(FFTW_LDOUBLE)
typedef long double R;
#  define X(name) FFTW_MANGLE_LONG_DOUBLE(name)
#else
typedef double R;
#  define X(name) FFTW_MANGLE_DOUBLE(name)
#endif

typedef R E;

/* duplicate some code from ifftw.h */

#ifdef FFTW_LDOUBLE
#  define K(x) ((E) x##L)
#else
#  define K(x) ((E) x)
#endif
#define DK(name, value) const E name = K(value)

/* FFTW 3 internal error function */
extern void X(assertion_failed)(const char *s, int line, const char *file);
#define CK(ex)                                           \
      (void)((ex) || (X(assertion_failed)(#ex, __LINE__, __FILE__), 0))

/* FFTW 3 internal trig functions (more accurate than cos/sin) */
#ifdef FFTW_LDOUBLE
typedef long double trigreal;
#else
typedef double trigreal;
#endif
extern trigreal X(cos2pi)(int m, int n);
extern trigreal X(sin2pi)(int m, int n);

struct X(mdct_plan_s) {
     int n;
     R scale;
     R *in, *out;
     R *buf1, *buf2;
     R *W, *W2;
     X(plan) p;
};

X(mdct_plan) X(plan_mdct)(int n, R scale, R *in, R *out, unsigned flags)
{
     X(mdct_plan) p;
     int n2 = n/2;
     fftw_r2r_kind kind = FFTW_R2HC;

     if (n % 2)
	  return 0;

     scale *= 0.5; /* FFTW's DCT-IV has an extra factor of 2 in front */

     p = (X(mdct_plan)) malloc(sizeof(struct X(mdct_plan_s)));
     CK(p);

     p->n = n;
     p->scale = scale;
     p->in = in;
     p->out = out;
     p->buf1 = (R *) X(malloc)(sizeof(R) * n);
     p->buf2 = (R *) X(malloc)(sizeof(R) * n);
     
     p->p = X(plan_many_r2r)(1, &n2, 2,
			     p->buf1, 0, 1, n2,
			     p->buf2, 0, 1, n2,
			     &kind, flags);
     if (!p->p) {
	  X(mdct_destroy_plan)(p);
	  return 0;
     }

     {
	  int j, mt, nt;
	  nt = 2*n;
	  mt = n/4 + 1;
	  p->W = (R *) X(malloc)(sizeof(R) * mt * 2);
	  for (j = 0; j < mt; ++j) {
	       p->W[2*j] = X(cos2pi)(j, nt) * scale; 
	       p->W[2*j+1] = X(sin2pi)(j, nt) * scale; 
	  }
     }
     
     {
	  int j, mt, nt;
	  nt = 8*n;
	  mt = n;
	  p->W2 = (R *) X(malloc)(sizeof(R) * mt);
	  for (j = 0; j < mt; j += 2) {
	       p->W2[j] = X(cos2pi)(j + 1, nt); 
	       p->W2[j+1] = X(sin2pi)(j + 1, nt); 
	  }
     }
     
     return p;
}

void X(mdct_destroy_plan)(X(mdct_plan) p)
{
     if (p) {
	  X(free)(p->W2);
	  X(free)(p->W);
	  X(free)(p->buf2);
	  X(free)(p->buf1);
	  if (p->p) X(destroy_plan)(p->p);
	  free(p);
     }
}

void X(execute_mdct)(X(mdct_plan) ego, R *I, R *O)
{
     int n = ego->n, n2 = n/2;
     R *buf1 = ego->buf1;
     R *buf2;
     R *W = ego->W;
     R *W2;
     R scale = ego->scale;
     int i;

     buf1[0] = (K(-2.0) * scale) * (I[3*n2 - 1] + I[3*n2]); /* 0 */
     buf1[n2] = (K(-2.0) * scale) * (I[n2] - I[n2-1]); /* n-1 */
     for (i = 1; i + i < n2; ++i) {
	  int k = i + i;
	  E a, b, a2, b2;
	  {
	       E u, v;
	       u = I[(3*n2-1) + k] + I[3*n2 - k]; /* k - 1 */
	       v = I[3*n2 + k] + I[(3*n2-1) - k]; /* k */
	       a = -(u + v);
	       b2 = v - u;
	  }
	  {
	       E u, v;
	       u = I[(n2-1) - k] - I[n2 + k]; /* n - k - 1 */
	       v = I[n2 - k] - I[(n2-1) + k]; /* n - k */
	       b = u + v;
	       a2 = u - v;
	  }
	  {
	       E wa, wb;
	       wa = W[2*i];
	       wb = W[2*i + 1];
	       {
		    E apb, amb;
		    apb = a + b;
		    amb = a - b;
		    buf1[i] = wa * amb + wb * apb; 
		    buf1[n2 - i] = wa * apb - wb * amb; 
	       }
	       {
		    E apb, amb;
		    apb = a2 + b2;
		    amb = a2 - b2;
		    buf1[n2 + i] = wa * amb + wb * apb; 
		    buf1[n - i] = wa * apb - wb * amb; 
	       }
	  }
     }
     if (i + i == n2) {
	  E u, v;
	  u = -(I[2*n - 1] + I[n]); /* n2 - 1 */
	  v = I[0] - I[n - 1]; /* n2 */
	  buf1[i] = (u + v) * (W[2*i] * K(2.0));
	  buf1[n - i] = (u - v) * (W[2*i] * K(2.0));
     }
     
     
     /* child plan: two r2hc's of size n/2 */
     X(execute_r2r)(ego->p, buf1, buf2 = ego->buf2);
     
     W2 = ego->W2;
     O = ego->out;
     { /* i == 0 case */
	  E wa, wb;
	  E a, b;
	  wa = W2[0]; /* cos */
	  wb = W2[1]; /* sin */
	  a = buf2[0];
	  b = buf2[n2];
	  O[0] = wa * a + wb * b;
	  O[n - 1] = wb * a - wa * b;
     }
     W2 += 2;
     for (i = 1; i + i < n2; ++i, W2 += 2) {
	  int k;
	  E u, v, u2, v2;
	  u = buf2[i];
	  v = buf2[n2 - i];
	  u2 = buf2[n2 + i];
	  v2 = buf2[n - i];
	  k = (i + i) - 1;
	  {
	       E wa, wb;
	       E a, b;
	       wa = W2[0]; /* cos */
	       wb = W2[1]; /* sin */
	       a = u - v;
	       b = v2 - u2;
	       O[k] = wa * a + wb * b;
	       O[n - 1 - k] = wb * a - wa * b;
	  }
	  ++k;
	  W2 += 2;
	  {
	       E wa, wb;
	       E a, b;
	       wa = W2[0]; /* cos */
	       wb = W2[1]; /* sin */
	       a = u + v;
	       b = u2 + v2;
	       O[k] = wa * a + wb * b;
	       O[n - 1 - k] = wb * a - wa * b;
	  }
     }
     if (i + i == n2) {
	  int k = (i + i) - 1;
	  E wa, wb;
	  E a, b;
	  wa = W2[0]; /* cos */
	  wb = W2[1]; /* sin */
	  a = buf2[i];
	  b = buf2[n2 + i];
	  O[k] = wa * a - wb * b;
	  O[n - 1 - k] = wb * a + wa * b;
     }
}
